#
# See the file "LICENSE" for the full license governing this code.
#
require 'puppet_x/enterprisemodules/weblogic/tree'
require 'puppet_x/enterprisemodules/weblogic/bean_path'

module Puppet_X
  module EnterpriseModules
    module WebLogic
      module Type
        # @private
        def self.included(parent)
          #
          # We use send here because in ruby 1.8.7, include is a private
          # method. In ruby 1.9 this has become public.
          # TODO: When we lose support for ruby 1.8.7 change this to more standard
          #
          parent.send(:include, EasyType)
          parent.send(:include, Puppet_X::EnterpriseModules::WebLogic::Access)
          parent.extend Puppet_X::EnterpriseModules::WebLogic::TitleParser
          parent.extend(ClassMethods)
        end

        PUPPET_META_ATTRIBUTES = [:name,
                                  :alias,
                                  :audit,
                                  :before,
                                  :loglevel,
                                  :noop,
                                  :notify,
                                  :require,
                                  :schedule,
                                  :stage,
                                  :subscribe,
                                  :tag,
                                  :provider].freeze


        def wls_config
          @wls_config ||= Puppet_X::EnterpriseModules::WebLogic::Tree.new(domain)
        end

        def pre_run_check
          fail "#{domain} is not a registered domain name." unless registered_domain?(domain)
        end


        def on_create
          Puppet.info "create #{name} "
          create_resource
          nil
        end

        def create_resource
          creation_path = BeanPath.new(bean_path.up)
          actual_creation_path = creation_path.subst(self.to_hash)
          fail "Parent resource #{actual_creation_path} doesn't exists." unless wls_config.exists?(actual_creation_path)
          wls_config.create(bean_type, actual_creation_path, self[key_name])
        end

        def destroy_resource
          wls_config.destroy(bean_type, bean_path.subst(to_hash), self[key_name])
        end

        def domain_name
          @domain_name ||= wls_config.get('/Name')
        end

        def to_hash
          hash = super
          hash = hash.merge( {:domain_name => domain_name})
          hash
        end

        def on_modify
          Puppet.info "modify #{name} "
          # Let the attributes do their work
          nil
        end

        def on_destroy
          Puppet.info "destroy #{name} "
          destroy_resource
          nil
        end

        def bean_type
          self.class.bean_type
        end

        def bean_path
          self.class.bean_path
        end

        def autorequire_when(type, type_name)
          return_value = []
          targets = self[:target] || []
          targets.each_with_index do |value, index|
            return_value << full_name(value) if self[:targettype][index] == type_name
          end
        end

        def key_name
          self.class.key_name
        end

        def registered_domain?(name)
          if catalog && catalog.resource('wls_setting', name)
            #
            # Look in the catalog and check if there is already a wls_setting for this resource. If
            # it exists, we are running a normal(full) catalog with wls_settings done.
            #
            true
          else
            #
            # If it doesn't exist. We are either running puppet resource or a mini manifest
            # expecting WLS to run. In order to run the wls_settings has to be saved and available
            #
            self.class.configuration.keys.include?(name)
          end
        end


        module ClassMethods
          class ExtraProps
            def initialize(parent)
              @parent = parent
            end

            def property(property_name)
              #
              # First load all extended properties like normal ones and then
              # create insync? method that will show an error when the attribute is used.
              #
              # Before we get the instances this method will be removed for enabled
              # extra properties.
              #
              @parent.property("extended_properties/#{property_name}")
              @parent.class_eval <<-EOD, __FILE__, __LINE__
                class #{property_name.to_s.capitalize}

                  def self.extended_attribute?
                    true
                  end

                  alias_method :original_insync?, :insync?

                  class <<self
                    alias_method :original_translate_to_resource, :translate_to_resource
                  end

                  def insync?(is)
                    fail "property '\#\{name\}' on '\#\{resource.name\}' is not enabled. Please enable it, or remove it."
                  end

                  def self.translate_to_resource(raw_resource, resource)
                    # This method returns nil
                    nil
                  end

                end

              EOD
            end

            def method_missing(*args)
              raise ArgumentError, "Invalid method called in extra properties block"
            end

          end

          def module_name
            'wls_config'
          end

          def execute_prefetch(resources, provider)
            resource_names_in_manifest = resources.keys
            name_parameter_class = paramclass(:name)
            get_raw_resources.each do |raw_resource|
              name = name_parameter_class.translate_to_resource(raw_resource)  # Get the full name
              next unless resource_names_in_manifest.include?(name)            # Skip if not in manifest
              specfied_properties = resources[name].to_hash.reject { |key, _value| PUPPET_META_ATTRIBUTES.include?(key) }.keys
              resources[name].provider = provider.map_raw_to_resource(raw_resource, specfied_properties)
            end
            resources
          end

          def extraproperties(&block)
            @extra_properties = ExtraProps.new(self)
            @extra_properties.instance_eval(&block)
          end

          def index_condition(&condition)
            @index_condition = condition
          end

          def get_raw_resources
            Puppet.info "index #{name} "
            raw_resources
          end

          def raw_resources
            enable_extended_methods
            configuration.keys.collect do |domain|
              wls_config = Puppet_X::EnterpriseModules::WebLogic::Tree.new(domain)
              domain_name = wls_config.get('/Name')
              resources_at(domain, bean_path, {:domain_name => domain_name}, [])
            end.flatten
          end

          def key_name
            (key_attributes - [:domain, :name]).first
          end

          def resources_at(domain, path, config, resources)
            wls_config = Puppet_X::EnterpriseModules::WebLogic::Tree.new(domain)
            resulting_path = path.subst(config)
            unless path.unassigned(resulting_path).empty?
              children = wls_config.children(resulting_path)
              children.each do | child|
                new_config = config.dup
                new_config[path.unassigned(resulting_path).first] = child
                resources_at( domain, path, new_config, resources)
              end
            else
              children = wls_config.children(resulting_path)
              #
              # No check if there is a callback
              #
              if @index_condition
                children = children.find_all {|c| @index_condition.call(c, "#{resulting_path}", wls_config)}
              end
              resources << children.collect {|m| [{key_name => m, :domain => domain}.merge(config)]}.flatten
            end
            resources = resources.flatten
            validate_resource_names(resources)
            resources
          end

          def validate_resource_names(resources)
            resources.each do | resource|
              resource.keys.map(&:to_s).grep(/name/i).each do |key|
                fail "WebLogic property: '#{key}' has value '#{resource[key.to_sym]}' which contains characters not support by wls_config module. Use weblogic names without a '\' and a `:`" unless resource[key.to_sym].scan(/\/|\:/).empty?
              end
            end
          end


          def enable_extended_methods
            enabled_extended_properties.each {|property|use_property(property)}
          end

          def enabled_extended_properties
            properties.collect{|p| p.name}.select{|p| enabled_property?(p)}
          end

          def enabled_property?(property)
            configuration.any? do |domain, config|
              enabled_properties = config.fetch('extra_properties') {[]}
              enabled_properties.any? do | entry|
                enabled_type, enabled_property = entry.split(':')
                case
                when !Regexp.new(enabled_type).match(name.to_s) then false
                when !Regexp.new(enabled_property).match(property.to_s) then false
                when !attrclass(property).respond_to?(:extended_attribute?) then false
                else true
                end
              end
            end
          end

          def use_property(property_name)
            class_eval <<-EOD, __FILE__, __LINE__
              class #{property_name.to_s.capitalize}

                 alias_method :insync?, :original_insync?

                 class <<self
                   alias_method :translate_to_resource, :original_translate_to_resource
                 end
              end

            EOD

          end

          def bean_type(bean_type = nil)
            if bean_type
              @bean_type = bean_type
            else
              @bean_type
            end
          end

          def bean_path(bean_path = nil)
            if bean_path
              @bean_path = BeanPath.new(bean_path)
            else
              case
              when @bean_path
                @bean_path
              when @bean_type
                @bean_path ||= BeanPath.new("/#{@bean_type}s")
              else
                raise(ArgumentError, 'no bean type or path specified') unless @bean_type || @bean_path
              end
            end
          end

        end
      end
    end
  end
end
