#
# See the file "LICENSE" for the full license governing this code.
#
require 'puppet_x/enterprisemodules/weblogic/functions'
require 'puppet_x/enterprisemodules/weblogic/types'

class Puppet::Parameter::WlsProperty < Puppet::Property
  include EasyType
  extend Puppet_X::EnterpriseModules::WebLogic::Functions

  def self.maps_to(path, &default_path)
    @wls_path = if path.is_a?(String)
                  [path]
                else
                  path
                end
    @default_path = default_path
  end

  def wls_config
    @wls_config ||= Puppet_X::EnterpriseModules::WebLogic::Tree.new(resource[:domain])
  end

  def self.wls_path
    if @wls_path
      @wls_path
    else
      fail "Internal Error. Property #{name} needs to be mapped with 'maps_to' function"
    end
  end

  def wls_path
    self.class.wls_path
  end

  def insync?(is)
    if should.to_s == 'absent' && is.to_s != 'absent'
      is.to_s == default_value.to_s
    elsif should.to_s == 'absent' && is.to_s == 'absent'
      true
    else
      is.to_s == should.to_s
    end
  end

  def default_value
    @default_value ||= wls_config.default_value(actual_path(resource.to_hash))
  end

  def self.array_separator(value = nil)
    if value
      @array_separator = value
    else
      @array_separator ||= ','
    end
  end

  def self.translate_to_resource(raw_resource, resource)
    my_resource = raw_resource.merge(resource) # Merge raw with what we got
    wls_config = Puppet_X::EnterpriseModules::WebLogic::Tree.new(my_resource[:domain])
    value = if is_folder?(actual_path(my_resource))
              wls_config.children(actual_path(my_resource), &@default_path)
            else
              wls_config.get(actual_path(my_resource), &@default_path)
    end
    if array_matching == :all && value.is_a?(String)
      value.split(array_separator)
    #
    # If the value is an Array but no array_matching is specified,
    # we are probably a reference
    #
    elsif array_matching != :all && value.is_a?(Array)
      value.first
    else
      value
    end
  end

  on_apply do
    apply_paths(value)
  end

  def apply_paths(value)
    actual_paths(resource.to_hash).each do |path|
      wls_config.set(path, value)
    end
  end

  def apply_paths_with_reference(parent, value)
    actual_paths(resource.to_hash).each do |path|
      wls_config.set_reference(path, parent, value)
    end
  end

  def apply_paths_with_references(parent, value)
    actual_paths(resource.to_hash).each do |path|
      wls_config.set_references(path, parent, value)
    end
  end

  def apply_paths_with_array(value)
    actual_paths(resource.to_hash).each do |path|
      wls_config.set_array(path, value)
    end
  end

  def apply_paths_with_property(value)
    actual_paths(resource.to_hash).each do |path|
      wls_config.set_property(path, value)
    end
  end

  def apply_paths_with_property_tree(value, property_type)
    actual_paths(resource.to_hash).each do |path|
      wls_config.set_property_tree(path, value, property_type)
    end
  end

  def apply_paths_with_target(value, type)
    actual_paths(resource.to_hash).each do |path|
      wls_config.set_target(path, value, type)
    end
  end

  def path_for(key)
    "#{type_path}/#{resource[key_name]}/#{key}" % resource.to_hash
  end

  def type_path
    self.class.type_path
  end

  def key_name
    self.class.key_name
  end

  def actual_paths(resource)
    self.class.actual_paths(resource)
  end

  def actual_path(resource)
    self.class.actual_path(resource)
  end

  private

  def self.is_folder?(path)
    path[-1, 1] == '/'
  end

  def self.actual_paths(resource)
    wls_path.collect do |p| 
      if p[0] == '/'  # Fullk path
        "#{p}" % resource.to_hash
      else
        "#{type_path}/#{resource[key_name]}/#{p}" % resource.to_hash
      end
    end
  end

  def self.actual_path(resource)
    actual_paths(resource).first
  end

  ##
  #
  # This magic returns the bean path specified for the type
  #
  def self.type_path
    type_class.bean_path
  end

  def self.key_name
    type_class.key_name
  end

  def self.type_class
    type_class_name = to_s.split('::')[0..2].join('::')
    class_from_string(type_class_name)
  end

  def self.class_from_string(str)
    str.split('::').inject(Object) do |mod, class_name|
      mod.const_get(class_name)
    end
  end
end
